function [m, k, fm, flag, varargout] = bisection(f, a, b, maxit, tol, mindx, info)

%BISECTION  Find zero of function in interval [a,b] with bisection method
% function [m, k, fm, flag, varargout] = bisection(f, a, b, maxit, tol, mindx, info)
%
% Stops if one of these conditions holds:  (with default value)
%   - |f(x)| < tol  (1e-6)
%   - relative change in x < mindx  (1e-3) (NB: set to 0 to enforce first condition)
%   - maxit number of iterations, both for optional first part and for main loop  (50)
%
% Revision date: May 26, 2016
% (C) Michiel Hochstenbach 2016

% Assign default values to unspecified parameters
if nargin < 2 || isempty(a)
  a = -1;
end
if nargin < 3 || isempty(b)
  b = 1;
end
if nargin < 4 || isempty(maxit)
  maxit = 50;    % 2^50 ~ 1e15
end
if nargin < 5 || isempty(tol)
  tol = 1e-6;    % |f(x)|
end
if nargin < 6 || isempty(mindx)
  mindx = 1e-3;  % Relative change in x
end
if nargin < 7 || isempty(info)
  info = 0;      % Info desired
end

if info
  fprintf('     x          f(x)       rel dx\n')
  fprintf('-----------------------------------\n')
end

% Take care that sign(f(a)) ~= sign(f(b))
try                     % Test if f has extra output arguments
  [fa, varargout] = f(a);
  extra = true;
catch
  fa = f(a);
  extra = false;
end
fb = f(b);
for k = 1:maxit
  if fa * fb < 0
    break;
  end
  b = a + (b-a)/2;      % More stable than (a+b)/2
  if info
    fprintf('Updating initial interval: end point %g\n', b)
  end
  fb = f(b);
end
if k > 1 && nargout < 3
  fprintf('Warning: had to update end point to %g\n', b);
end

% Main loop
for k = 1:maxit
  m  = a + (b-a)/2;
  if extra
    [fm, varargout] = f(m);
  else
    fm = f(m);
  end
  r = abs(a-b) / abs(a);
  if info          % Info
    fprintf('%10.3e  %10.3e  %10.3e\n', m, fm, r)
  end
  if r < mindx || abs(fm) < tol
    break;
  end
  if fm * fb < 0
    a  = m;
  else
    b  = m;
    fb = fm;
  end
end

% Info
flag = 0;
if k == maxit
  if nargout < 3
    fprintf('Warning: bisection.m did not converge to tolerance %g in %d steps\n', tol, maxit);
  else
    flag = 1;
  end
end

if exist('varargout', 'var')
  varargout = {varargout};
end
