function [low, up, iter] = condprob(A, Ainv, v1, epsilon, ratio, info)

%CONDPROB  Lower and probabilistic upper bound for 2-norm condition number kappa(A)
% function [low, up, iter] = condprob(A, Ainv, v1, epsilon, ratio, info)
%
% Input  (with default value)
%   A         : (real or complex) nonsingular matrix
%   v1        : Initial vector  (randn1)  (NB: provide a v1 if you want repeatable experiments)
%   epsilon   : Desired probabilistic upper bounds are with probability >= 1-epsilon  (0.01)
%   ratio     : Desired upper bound for ratio "probabilistic up / low"                (1.01)
%   info      : Info desired  (NB: tip: try option 1)                                 (0)
%                 0 gives no information
%                 1 gives short table at the end for >= 95%, 99%, and 99.9% probability levels
%                 2 also gives intermediate output (NB: slows down the code somewhat)
%
% Output
%   low       : Lower bound computed with Lanczos bidiagonalization
%   up        : Probabilistic upper bound computed with Lanczos bidiagonalization polynomials
%   iter      : Number of iterations (automatically selected) 
%
% Examples:
%   v1 = randn(size(A,2),1);  (useful to get repeatable results)
%   [low, up] = condprob(A, v1);
%      Gives lower, and >= 99% upper bound for ||A||, and in addition up/low <= 1.01
%   [low, up] = condprob(A, v1, 0.001);
%      Gives lower, and >= 99.9% upper bound for ||A||, and in addition up/low <= 1.01
%   [low, up] = condprob(A, v1, [], 1.1);
%      Gives lower, and >= 99% upper bound for ||A||, and in addition up/low <= 1.1
%   [low, up] = condprob(A, v1, [], [], 1);
%      Gives lower, and >= 99% upper bound for ||A||, and in addition up/low <= 1.01
%      Also, at the end a short table is given for 95%, 99%, and 99.9%
%   [low, up] = condprob(A, v1, [], [], 2);
%      Gives lower, and >= 99% upper bound for ||A||, and in addition up/low <= 1.01
%      Also, at the end a short table is given for 95%, 99%, and 99.9%, and intermediate results are printed
%
% Reference:
%   S.W. Gaaf, M.E. Hochstenbach
%   Probabilistic bounds for the matrix condition number with extended Lanczos bidiagonalization
%   SIAM J. Sci. Comp. 37(5), pp. S581-S601, 2015
%
%
% See also NORMPROB, LANPROB, LANPROB_DELTA
%
% Revision date: May 29, 2016
% (C) Sarah Gaaf, Michiel Hochstenbach 2016

% Parameters: small since inverse may be expensive
initial_dim =  2;   % Start with (A'A)^{-2}v, (A'A)^{-1}v, v, (A'A)v, (A'A)^2v
delta_dim   =  1;   % Continue with (A'A)^k v, (A'A)^{-k}v

% Problem size
n = size(A,2);

% Default values
if nargin < 3 || isempty(v1)
  v1 = randn1(n);   % Initial vector for Krylov space
end
if nargin < 4 || isempty(epsilon)
  epsilon = 1e-2;   % Upper bound holds with probability >= 99%
end
epsilon = epsilon / 2;  % Upper bound holds with probability >= 1-2*epsilon
if nargin < 5 || isempty(ratio)
  ratio = 1.01;     % Upper bound for ratio: Probabilistic upper bound / lower bound
end
if nargin < 6 || isempty(info)
  info = 0;         % Default: no information
end

if isempty(Ainv)
  Ainv = lu0(A);    % LU-decomposition
end

% Probabilistic constant; see paper for more info
delta_inv = 1 / lanprob_delta(n, epsilon);

if info > 1         % Provide intermediate output 
  fprintf('   k     low     up   ratio\n');
  fprintf('----------------------------\n');
end

% Lower and probabilistic upper bounds with Lanczos bidiagonalization polynomials
% Initial extended Krylov subspace K_k(A'A, v1) + K_k(inv(A'A), v1)
% Last input argument should be even
[U, V, C, D, H, K] = krylov_ata_extended(A, Ainv, v1, 2*initial_dim);

for j = 1:100
  % ||H|| <= ||A|| and ||K|| >= ||A^{-1}||
  theta  = norm(H(1:end-1,1:end-1));
  xi     = norm(K(1:end-1,:));
  theta2 = norm(H);
  xi2    = norm(K);
  smax_low = theta2;
  smin_up  = 1/xi2;

  % Compute probabilistic upper bound
  [smax_up_prob, smin_low_prob, flag] = prob_bounds(C, D, theta, xi, delta_inv);
  up(j)  = smax_up_prob / smin_low_prob;
  low(j) = smax_low / smin_up;
  if j == 1
    iter = initial_dim;
  else
    iter(j) = iter(j-1) + delta_dim;
  end
  if flag
    r = NaN;        % No bound computed by bisection (should be rare, and only for low dimensions)
  else
    r = up(j) / low(j);
  end
  if (info == 2) || (info > 2 && ~mod(j, info))  % Display info
    fprintf('%5d  %5.3g   %5.3g  %5.3g   %5.3g  %5.3g   %5.3g  %.3f\n', size(U,2), ...
      smax_low, smax_up_prob, smin_low_prob, smin_up, low(j), up(j), r);
  end
  if ~flag && (up(j) <= ratio * low(j))
    break           % Success
  end

  % Expand extended Krylov space; last input argument should be even
  [U, V, C, D, H, K] = krylov_ata_extended_expand(A, Ainv, U, V, C, D, H, K, 2*delta_dim);
end

% Comment these 3 lines if convergence plot is desired
up  = up(j);
low = low(j);
iter = iter(j);

if info        % Print short table for 95%, 99%, 99.9%
  fprintf('Lower bound for k(A)\n')
  fprintf('          %6.4g\n', low(end))
  fprintf('Probabilistic upper bound for k(A) (with ratio up/low) \n')
  fprintf('          %6.4g\n', up(end))
  [smax95,  smin95]  = prob_bounds(C, D, theta, xi, 1 / lanprob_delta(n, 0.025));
  [smax99,  smin99]  = prob_bounds(C, D, theta, xi, 1 / lanprob_delta(n, 0.005));
  [smax999, smin999] = prob_bounds(C, D, theta, xi, 1 / lanprob_delta(n, 0.0005));
  fprintf('  95%%  :  %4.4g   (%5.3f)\n', smax95/smin95,   smax95/(smin95*low(end)))
  fprintf('  99%%  :  %4.4g   (%5.3f)\n', smax99/smin99,   smax99/(smin99*low(end)))
  fprintf('  99.9%%:  %4.4g   (%5.3f)\n', smax999/smin999, smax999/(smin999*low(end)))
  fprintf('Matrix-vector products: %d with each of A, A^T, A^{-1}, and A^{-T}\n', j)
end


function [smax_up_prob, smin_low_prob, flag] = prob_bounds(C, D, theta, xi, delta_inv)
pplus = @(x) bidiag_ext_poly(x, C, D, 1);
pmin  = @(x) bidiag_ext_poly(x, C, D, 2);
up0  = 2*theta;
while abs(pplus(up0)) < delta_inv
  up0 = 2*up0;
end
low0 = 2*xi;
while abs(pmin(1/low0)) < delta_inv
  low0 = 2*low0;
end
[smax_up_prob,  m1] = bisection(@(x) pplus(x) - pplus(theta) - sign(pplus(2*theta)) * delta_inv, theta, up0, 100);
[smin_low_prob, m2] = bisection(@(x) pmin(1/x)- pmin(1/xi)   - sign(pmin(1/(2*xi))) * delta_inv, xi, low0, 100);
smin_low_prob = 1 / smin_low_prob;
if m1 == 100 || m2 == 100
  flag = 1;  % No bound computed by bisection (should be rare, and only for low dimensions)
else
  flag = 0;
end


function p = bidiag_ext_poly(x, C, D, out)

%BIDIAG_EXT_POLY  Evaluates the extended bidiagonalization polynomials
% function p = bidiag_ext_poly(x, C, D, out)
%   u_{k+1} = p_k(AA^*) Av_1
%   v_{k+1} = q_k(A^*A) v_1
% out indicates which is of interest

k = length(C);
q_plus = x / C(1,1);
p_plus = (x*q_plus - D(1,1))  / D(2,1);
q_min  = ((1/x)*p_plus -C(1,2)*q_plus) / C(2,2);
p_min  = ((1/x)*q_min - D(2,2)*p_plus - D(1,2)) / D(3,2);

for j = 3:2:k-1
  q_plus = (x*p_min - C(2,j)*q_min - C(1,j)*q_plus) / C(3,j);
  p_plus = (x*q_plus - D(2,j)*p_min - D(1,j)*p_plus) / D(3,j);
  q_min  = ((1/x)*p_plus - C(2,j+1)*q_plus - C(1,j+1)*q_min) / C(3,j+1);
  p_min  = ((1/x)*q_min - D(2,j+1)*p_plus - D(1,j+1)*p_min) / D(3,j+1) ;      
end

if mod(k,2) == 1    
  q_plus = (x*p_min - C(2,k)*q_min - C(1,k)*q_plus) / C(3,k);
  p_plus = (x*q_plus - D(2,k)*p_min - D(1,k)*p_plus) / D(3,k);
end

switch out
  case 1, p = p_plus;
  case 2, p = p_min;
end
