function [U, V, C1, D1, H, K] = krylov_ata_extended(A, Ainv, b, k)

%KRYLOV_ATA_EXTENDED  Construct orthonormal basis for extended bidiagonalization subspace
% [U, V, C1, D1, H, K] = krylov_ata_extended(A, Ainv, b, k)
%
% Ainv(x,0) should carry out A\x
% Ainv(x,1) should carry out A'\x
%
% See also KRYLOV_ATA_EXTENDED_EXPAND
%
% Revision date: November 14, 2014
% (C) Sarah Gaaf, Michiel Hochstenbach 2016

if nargin < 3 || isempty(b)
  b = randn(size(A,2), 1);
end
if nargin < 4 || isempty(k)
  k = 4;
end

U = [];                                         % 
V = b / sqrt(b'*b);

for j = 1:k
  if mod(j,2) == 1
    if j == 1
      [U(:,j), C(1,j)]     = rgs(mv(A,V(:,j),0), U);            % u = Av
      [V(:,j+1), D(1:2,j)] = rgs(mv(A,U(:,j),1), V);            % v = A'u
      C1(1,j) = C(1,j);
      D1(1:2,j) = D(1:2,j);
    else
      [U(:,j), C(1:j,j)]     = rgs(mv(A,V(:,j),0), U);          % u = Av
      [V(:,j+1), D(1:j+1,j)] = rgs(mv(A,U(:,j),1), V);          % v = A'u
      C1(1:3,j) = C(j-2:j,j);  
      D1(1:3,j) = D(j-1:j+1,j);
    end
  else
    if j == 2
      [U(:,j), C(1:2,j)]   = rgs(mv(Ainv,V(:,j),1), U);         % u = A'\v  
      [V(:,j+1), D(1:3,j)] = rgs(mv(Ainv,U(:,j),0), V);         % v = A\u
      C1(1:2,j) = C(1:2,j);
      D1(1:3,j) = D(1:3,j);
    else
      [U(:,j), C(1:j,j)]     = rgs(mv(Ainv,V(:,j),1), U);       % u = A'\v   
      [V(:,j+1), D(1:j+1,j)] = rgs(mv(Ainv,U(:,j),0), V);       % v = A\u
      C1(1:3,j) = C(j-2:j,j);
      D1(1:3,j) = D(j-1:j+1,j);
    end
  end
end

%-------------------------------------------------------------------------
% Construction of the matrices H and K
%-------------------------------------------------------------------------
H = zeros(k,k);
K = zeros(k+1,k);
for j = 1:2:k
  H(j,:) = D(1:k,j)';
  K(j,j) = 1/C(j,j);
end
for j = 2:2:k
  H(j,j) = 1/C(j,j);
  K(:,j) = D(:,j);
end
