function [V, alpha, beta] = krylov_sym(A, b, k, full, reortho)

%KRYLOV_SYM  Construct orthonormal basis for Krylov subspace Hermitian mmatrices
% function [V, alpha, beta] = krylov_sym(A, b, k, full, reortho)
%
% Out: AV_k = V_{k+1} T_{k+1,k}
% alpha    diagonal elements of T_k
% beta     subdiagonal and superdiagonal elements of T_k
%
% k        dimension of Krylov space
% full     0: basis for V_k,     length(beta) = k
%          1: basis for V_{k+1}, length(beta) = k+1
% reortho  0: no reorthogonalization (often not recommended)
%          1: reorthogonalization (default)
%
% See also KRYLOV_SCHUR_SYM, KRYLOV_SYM_EXPAND, KRYLOV, KRYLOV_ATA
%
% Revision date: May 29, 2016
% (C) Michiel Hochstenbach 2016

if nargin < 2 || isempty(b)
  if ~isnumeric(A)
    error('Need a starting vector if A is a function handle')
  end
  b = randn1(size(A,2));
end
if nargin < 3 || isempty(k)
  k = 10;
end
if nargin < 4 || isempty(full)
  full = 1;
end
if nargin < 5 || isempty(reortho)
  reortho = 1;
end

normb = sqrt(b'*b);

% Preallocation
alpha = zeros(1,k);
beta  = zeros(1,k-1);
if reortho
  V      = zeros(length(b), k);
  V(:,1) = b * (1 / normb);
else
  v = b * (1 / normb);
end

for j = 1:k
  if j == k && ~full && nargout < 2
    % All maxit necessary vectors have been computed
    return
  end
  if reortho
    r = mv(A, V(:,j));
  else
    r = mv(A, v);
  end
  if j > 1
    if reortho
      r = r - beta(j-1)*V(:,j-1);
    else
      r = r - beta(j-1)*v0;
    end
  end
  if reortho
    alpha(j) = V(:,j)'*r;
    r = r - alpha(j)*V(:,j);
    % Reorthogonalization
    r = r - V*(V'*r);
  else
    alpha(j) = v'*r;
    r = r - alpha(j)*v;
  end
  if j < k || full
    beta(j) = sqrt(r'*r);
    if reortho
      V(:,j+1) = r * (1 / beta(j));
    else
      v0 = v;
      v  = r * (1 / beta(j));
    end
  end
end

if ~reortho
  V = [v0 v];
end
