function [low, up, lmin, k] = lanprob(A, v1, epsilon, ratio, info)

%LANPROB  Lower bound and and probabilistic upper bound for the largest eigenvalue
%         of a symmetric matrix with the Lanczos method
%
% function [low, up, lmin, k] = lanprob(A, v1, epsilon, ratio, info)
%
% Input
%   v1      : Initial vector
%   epsilon : Desired probabilistic upper bounds are with probability >= 1-epsilon
%   ratio   : Desired upper bound for ratio (up-low) / (low-lmin)
%               where lmin is the approximation to the minimum eigenvalue
%               as computed during the process
%   info    : Intermediate info desired (slows down the code)
%
% Output
%   low     : Guaranteed lower bound computed with Lanczos
%   up      : Probabilistic upper bound computed with Lanczos polynomials
%   lmin    : Approximate minimum eigenvalue
%   k       : Dimension of Krylov space (automatically selected) 
%
% Reference:
%   J.L.M. van Dorsselaer, M.E. Hochstenbach, H.A. van der Vorst
%     Computing probabilistic bounds for extreme eigenvalues
%     of symmetric matrices with the Lanczos method,
%     SIAM J. on Matrix Anal. Appl. 22(3), pp. 837-852, 2000
%
% See also NORMPROB, CONDPROB, LANPROB_DELTA, KRYLOV_SYM, KRYLOV_SYM_EXPAND
%
% Revision date: May 29, 2016
% (C) Michiel Hochstenbach 2016

n = size(A,2);
initial_dim = 5;
delta_dim = 6;

% Default values
if nargin < 2 || isempty(v1)
  v1 = randn(n,1); % Initial vector for Krylov space
end
if nargin < 3 || isempty(epsilon)
  epsilon = 1e-2;  % Reliability probabilistic upper bound >= 99%
end
if nargin < 4 || isempty(ratio)
  ratio = 0.01;    % Upper bound for ratio: (up-low) / (low-lmin)
end
if nargin < 5 || isempty(info)
  info = 0;        % Info desired
end

% First very rough upper bound: 8 * ||A||_1
if isnumeric(A)
  up0 = 8 * norm(A, 'inf');
else
  up0 = 8 * normest1(@(flag, x) mv(A, x, flag));
end

% Probabilistic constant
delta_inv = 1 / lanprob_delta(n, epsilon);

if info > 1
  fprintf('   k      low         up      lmin    ratio\n');
  fprintf('--------------------------------------------\n');
end

% Lower and probabilistic upper bounds with Lanczos polynomials
% Initial subspace
[V, alpha, beta] = krylov_sym(A, v1, initial_dim, 1);

for j = 1:1000
  % Largest Ritz value is guaranteed lower bound
  beta0 = beta(1:length(beta)-1);
  T = diag(alpha) + diag(beta0,1) + diag(beta0,-1);
  [~,D] = eig(T);
  D = diag(D);
  low  = max(D);    % Lower bound for lambda_max
  lmin = min(D);    % Approximation (and upper bound) for lambda_min 

  % Compute probabilistic upper bound
  [up, m] = bisection(@(x) lanpoly(x, alpha, beta)-delta_inv, low, up0, 100, 1e-3);
  if m < 100
    r = (up-low) / (low-lmin);
  else
    r = NaN;        % No bound computed by bisection
  end
  if (info == 2) || (info > 2 && ~mod(k, info))  % Display info
    fprintf('%5d  %8.3g   %8.3g  %8.3g  %.3f\n', length(alpha), low, up, lmin, r);
  end
  if r <= ratio
    k = size(V,2);
    break           % Success
  end
  up0 = up;         % Store previous upper bound for next iteration

  % Expand Krylov space
  if (m < 100) && (r < 1.003 * ratio)
    % Close to convergence: carry out fewer iterations
    dim = ceil(delta_dim / 2);
  else
    dim = delta_dim;
  end
  [V, alpha, beta] = krylov_sym_expand(A, V, alpha, beta, dim);
end

if info        % Print short table for 95%, 99%, 99.9%
  fprintf('Lower bound for lambda_max\n')
  fprintf('           %4.4g\n', low)
  fprintf('Probabilistic upper bound for ||A|| (with ratio (up-low)/(low-lmin))\n')
  delta95  = 1 / lanprob_delta(n, 0.05);
  delta99  = 1 / lanprob_delta(n, 0.01);
  delta999 = 1 / lanprob_delta(n, 0.001);
  up0 = up + (up-lmin) / 2;
  up999 = bisection(@(x) lanpoly(x, alpha, beta) - delta999, low, up0,  [], [], 1e-4);
  up99  = bisection(@(x) lanpoly(x, alpha, beta) - delta99,  low, up999,[], [], 1e-4);
  up95  = bisection(@(x) lanpoly(x, alpha, beta) - delta95,  low, up99, [], [], 1e-4);
  fprintf('  95%%  :  %4.4g   (%5.3f)\n', up95,  (up95-low) /(low-lmin))
  fprintf('  99%%  :  %4.4g   (%5.3f)\n', up99,  (up99-low) /(low-lmin))
  fprintf('  99.9%%:  %4.4g   (%5.3f)\n', up999, (up999-low)/(low-lmin))
  fprintf('Matrix-vector products: %d with A\n', k)
end


function p = lanpoly(x, alpha, beta)
%LANPOLY  Lanczos polynomial evaluated at x
% function p = lanpoly(x, alpha, beta)
k  = length(alpha);
p0 = 1;
p1 = (x-alpha(1)) / beta(1);
p  = k*p1 + 1-k;  % In case k=0 or k=1
for j = 2:k
  p  = ((x-alpha(j)).*p1 - beta(j-1)*p0) / beta(j);
  p0 = p1;
  p1 = p;
end


function [V, alpha, beta] = krylov_sym_expand(A, V, alpha, beta, m)
%KRYLOV_SYM_EXPAND  Expand orthonormal basis for Krylov subspace K_k(A,b) to K_{k+m}(A,b), A'=A
% function [V, alpha, beta] = krylov_sym_expand(A, V, alpha, beta, m)
k = length(alpha)+1;
V = [V zeros(size(V,1), m)];   % Preallocation
alpha = [alpha zeros(1,m)];
beta  = [beta  zeros(1,m)];
for j = k:(m+k-1)
  r = mv(A, V(:,j));
  r = r - beta(j-1)*V(:,j-1);
  alpha(j) = V(:,j)'*r;
  r = r - alpha(j)*V(:,j);
  r = r - V*(V'*r);            % Reorthogonalization
  beta(j) = sqrt(r'*r);
  V(:,j+1) = r * (1 / beta(j));
end
