function [low, up, smin, k] = normprob(A, v1, epsilon, ratio, info)

%NORMPROB  Lower and probabilistic upper bound for ||A||_2
% function [low, up, smin, k] = normprob(A, v1, epsilon, ratio, info)
%
% Input  (with default value)
%   A       : (real or complex) (square or nonsquare) matrix, in explicit form
%               or function of 2 arguments @(x,flag) such that A(x,0) = A*x and A(x,1) = A'*x
%               In the latter case, a starting vector v1 of the correct length has to be provided
%   v1      : Initial vector  (randn)   (NB: provide a v1 if you want repeatable experiments)
%   epsilon : Desired probabilistic upper bounds are with probability >= 1-epsilon  (0.01)
%   ratio   : Desired upper bound for ratio "probabilistic up" / low  (1.01)
%   info    : Info desired (0)  (NB: tip: try option 1)
%               0 gives no info
%               1 gives short table at the end for >= 95%, 99%, and 99.9% probability levels
%               2 also gives intermediate info (slows down the code)
%
% Output
%   low     : Lower bound computed with Lanczos bidiagonalization
%   up      : Probabilistic upper bound computed with Lanczos bidiagonalization polynomials
%   smin    : Upper bound to sigma_min
%   k       : Dimension of Krylov space (automatically selected) 
%
% Examples:
%   v1 = randn(size(A,2),1);  (useful to get repeatable results)
%   [low, up] = normprob(A, v1);
%      Gives lower, and >= 99% upper bound for ||A||, and in addition up/low <= 1.01
%   [low, up] = normprob(A, v1, 0.001);
%      Gives lower, and >= 99.9% upper bound for ||A||, and in addition up/low <= 1.01
%   [low, up] = normprob(A, v1, [], 1.1);
%      Gives lower, and >= 99% upper bound for ||A||, and in addition up/low <= 1.1
%   [low, up] = normprob(A, v1, [], [], 1);
%      Gives lower, and >= 99% upper bound for ||A||, and in addition up/low <= 1.01
%      Also, at the end a short table is given for 95%, 99%, and 99.9%
%   [low, up] = normprob(A, v1, [], [], 2);
%      Gives lower, and >= 99% upper bound for ||A||, and in addition up/low <= 1.01
%      Also, at the end a short table is given for 95%, 99%, and 99.9%, and intermediate results are printed
%
% Reference:
%   M.E. Hochstenbach
%    Probabilistic upper bounds for the matrix two-norm
%    J. Sci. Comp. 57 (3), pp. 464-476, 2013
%   
% See also LANPROB, CONDPROB
%
% Revision date: May 29, 2016
% (C) Michiel Hochstenbach 2016

% Parameters
initial_dim = 5;
delta_dim   = 6;

% 0 no reorthogonalization (not recommended: faster, at the expense of the accuracy of the results)
% 1 reorthogonalization of the V vectors (slower but more accurate)
% 2 reorthogonalization of both the U and V vectors (slower but more accurate)
reortho = 2;     % Reorthogonalization

% Problem size
if isnumeric(A)
  n = size(A,2);
else
  if nargin < 2 || isempty(v1)
    error('Need a starting vector if A is a function handle')
  else
    n = length(v1);  
  end
end

% Default values
if nargin < 2 || isempty(v1)
  v1 = randn(n,1); % Initial vector for Krylov space
end
if nargin < 3 || isempty(epsilon)
  epsilon = 1e-2;  % Reliability >= 99%
end
if nargin < 4 || isempty(ratio)
  ratio = 1.01;    % Upper bound for ratio: Probabilistic upper bound / lower bound
end
if nargin < 5 || isempty(info)
  info = 0;        % Info desired
end

up0 = NaN;

% Probabilistic constant; see paper for more info
delta_inv = 1 / lanprob_delta(n, epsilon);

if info > 1   % Provide intermediate info
  fprintf('   k     low     up   ratio\n');
  fprintf('----------------------------\n');
end

% Lower and probabilistic upper bounds with Lanczos bidiagonalization polynomials
% Initial Krylov subspace K_k(A'A, v1)
[U, V, alpha, beta] = krylov_ata(A, v1, initial_dim, 1, reortho);

for j = 1:1000
  % Largest Ritz value is lower bound
  k = length(alpha);
  B = diag(alpha) + diag(beta(1:k-1),1);
  B(k,k+1) = beta(k);
  S = svd(B);
  low = S(1);
  smin = S(end);

  % Compute probabilistic upper bound
  if isnan(up0)     % bisection needs an end of interval for the first call
    up0  = 2*low;
    while abs(up0.* bidiagpoly(up0.^2, alpha, beta)) < delta_inv
      up0 = 2*up0;
    end
  end
  [up, m] = bisection(@(x) x.* bidiagpoly(x.^2, alpha, beta) - delta_inv, low, up0, [], [], 1e-4);
  if m < 100
    r = up / low;
  else
    r = NaN;        % No bound computed by bisection (should be rare, and only for low dimensions)
  end
  if (info == 2) || (info > 2 && ~mod(j, info))  % Display info
    fprintf('%5d  %5.3g   %5.3g  %.3f\n', length(alpha), low, up, r);
  end
  if (m < 100) && (up <= ratio * low)
    break           % Success
  end
  up0 = up;         % Store previous upper bound for next iteration

  % Expand Krylov space
  if r < 1.003 * ratio
    % Close to convergence: carry out fewer iterations
    dim = ceil(delta_dim / 2);
  else
    dim = delta_dim;
  end
  [U, V, alpha, beta] = krylov_ata_expand(A, U, V, alpha, beta, dim, reortho);
end

if info        % Print short table for 95%, 99%, 99.9%
  fprintf('Lower bound for ||A||\n')
  fprintf('          %6.4g\n', low)
  fprintf('Probabilistic upper bound for ||A|| (with ratio up/low) \n')
  delta95  = 1 / lanprob_delta(n, 0.05);
  delta99  = 1 / lanprob_delta(n, 0.01);
  delta999 = 1 / lanprob_delta(n, 0.001);
  up999 = bisection(@(x) x.* bidiagpoly(x.^2, alpha, beta) - delta999, low, 1.5*up, [], [], 1e-4);
  up99  = bisection(@(x) x.* bidiagpoly(x.^2, alpha, beta) - delta99,  low, up999,  [], [], 1e-4);
  up95  = bisection(@(x) x.* bidiagpoly(x.^2, alpha, beta) - delta95,  low, up99,   [], [], 1e-4);
  fprintf('  95%%  :  %4.4g   (%5.3f)\n', up95,  up95/low)
  fprintf('  99%%  :  %4.4g   (%5.3f)\n', up99,  up99/low)
  fprintf('  99.9%%:  %4.4g   (%5.3f)\n', up999, up999/low)
  fprintf('Matrix-vector products: %d with A and %d with A''\n', k, k)
end


function p = bidiagpoly(x, alpha, beta)
%BIDIAG_POLY  Evaluates the bidiagonalization polynomials
% function p = bidiagpoly(x, alpha, beta)
%   u_{k+1} = p_k(AA^*) Av_1
%   v_{k+1} = q_k(A^*A) v_1
p0  = 1 / alpha(1);
q0  = 1;
p   = p0;
q   = q0;
for k = 1:length(alpha)-1
  q  = (x.*p0 - alpha(k)*q0) / beta(k);
  p  = (q - beta(k)*p0) / alpha(k+1);
  p0 = p;
  q0 = q;
end


function [U, V, alpha, beta] = krylov_ata_expand(A, U, V, alpha, beta, k, reortho)
%KRYLOV_ATA_EXPAND  Expand orthonormal bases for Lanczos bidiagonalization spaces K_k(A'A, v) and K_k(AA', Av)
% function [U, V, alpha, beta] = krylov_ata_expand(A, U, V, alpha, beta, k)
m = length(alpha)+1;
V = [V zeros(size(V,1), k)];             % Preallocation
U = [U zeros(size(U,1), k)];
alpha = [alpha zeros(1,k)];
beta  = [beta  zeros(1,k)];
for j = m:(k+m-1)
  r = mv(A, V(:,j), 0);
  r = r - beta(j-1)*U(:,j-1);
  if reortho > 1
    r = r - U(:,1:j-1)*(U(:,1:j-1)'*r);  % Reorthogonalization
  end
  alpha(j) = sqrt(r'*r);
  U(:,j) = r * (1 / alpha(j));
  r = mv(A, U(:,j), 1);
  r = r - alpha(j)*V(:,j);
  if reortho
    r = r - V(:,1:j)*(V(:,1:j)'*r);      % Reorthogonalization
  end
  beta(j) = sqrt(r'*r);
  V(:,j+1) = r * (1 / beta(j));
end
